/*
 * $Id: dtrie.c 5108 2008-10-22 14:20:37Z henningw $
 *
 * Copyright (C) 2008 1&1 Internet AG
 *
 * This file is part of Kamailio, a free SIP server.
 *
 * Kamailio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * Kamailio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file
 * \brief Trie datastructure with utility functions
 *
 * Provides a generic trie datastructure and utility functions to
 * initialize and manage individual nodes. Its optimized towards
 * the usecase of a matching tree that contains only digits, e.g.
 * for LCR or blacklist modules. The trie is setup in shared memory.
 * - Module: \ref carrierroute
 * - Module: \ref userblacklist
 */


#include "dtrie.h"

#include <assert.h>

#include "../dprint.h"
#include "../mem/shm_mem.h"
#include "../mem/mem.h"


struct dtrie_node_t *dtrie_init(void)
{
	struct dtrie_node_t *root;

	root = shm_malloc(sizeof(struct dtrie_node_t));
	if (root == NULL) {
		SHM_MEM_ERROR;
		return NULL;
	}
	LM_DBG("allocate %lu bytes for root at %p", (long unsigned)sizeof(struct dtrie_node_t), root);
	memset(root, 0, sizeof(struct dtrie_node_t));

	return root;
}


void dtrie_delete(struct dtrie_node_t *root, struct dtrie_node_t *node, dt_delete_func_t delete_payload)
{
	unsigned int i;
	if (node==NULL) return;

	for (i=0; i<10; i++) {
		dtrie_delete(root, node->child[i], delete_payload);
		node->child[i] = NULL;
	}

	if (node != root) {
		LM_DBG("free node at %p", node);
		if (delete_payload) delete_payload(node->data);
		shm_free(node);
	}
}


void dtrie_destroy(struct dtrie_node_t **root, dt_delete_func_t delete_payload)
{
	if ((root!=NULL) && (*root!=NULL)) {
		dtrie_delete(*root, *root, delete_payload);
		LM_DBG("free root at %p", root);
		shm_free(*root);
		*root = NULL;
	}
}


void dtrie_clear(struct dtrie_node_t *root, dt_delete_func_t delete_payload)
{
	dtrie_delete(root, root, delete_payload);
	memset(root, 0, sizeof(struct dtrie_node_t));
}


int dtrie_insert(struct dtrie_node_t *root, const char *number, const unsigned int numberlen, void *data)
{
	struct dtrie_node_t *node = root;
	unsigned int digit, i=0;

	while (i<numberlen) {
		digit = number[i] - '0';
		if (digit>9) {
			LM_ERR("cannot insert non-numerical number\n");
			return -1;
		}
		if (node->child[digit] == NULL) {
			node->child[digit] = shm_malloc(sizeof(struct dtrie_node_t));
			assert(node->child[digit] != NULL);
			LM_DBG("allocate %lu bytes for node at %p", (long unsigned)sizeof(struct dtrie_node_t), node->child[digit]);
			memset(node->child[digit], 0, sizeof(struct dtrie_node_t));
		}
		node = node->child[digit];

		i++;
	}

	node->data = data;
	return 0;
}


unsigned int dtrie_size(const struct dtrie_node_t *root)
{
	unsigned int i, sum = 0;

	if (root == NULL) return 0;

	for (i=0; i<10; i++) {
		sum += dtrie_size(root->child[i]);
	}

	return sum+1;
}


unsigned int dtrie_loaded_nodes(const struct dtrie_node_t *root)
{
	unsigned int i, sum = 0;

	if (root == NULL) return 0;

	for (i=0; i<10; i++) {
		sum += dtrie_loaded_nodes(root->child[i]);
	}

	if (root->data != NULL) sum++;

	return sum;
}


unsigned int dtrie_leaves(const struct dtrie_node_t *root)
{
	unsigned int i, sum = 0, leaf = 1;

	for (i=0; i<10; i++) {
		if (root->child[i]) {
			sum += dtrie_leaves(root->child[i]);
			leaf = 0;
		}
	}

	return sum+leaf;
}


void **dtrie_longest_match(struct dtrie_node_t *root, const char *number, const unsigned int numberlen, int *nmatchptr)
{
	struct dtrie_node_t *node = root;
	unsigned digit, i = 0;
	void **ret = NULL;

	if (nmatchptr) *nmatchptr=-1;
	if (node->data != NULL) {
		if (nmatchptr) *nmatchptr=0;
		ret = &node->data;
	}
	while (i<numberlen) {
		digit = number[i] - '0';
		if (digit>9) return ret;
		if (node->child[digit] == NULL) return ret;
		node = node->child[digit];
		i++;
		if (node->data != NULL) {
			if (nmatchptr) *nmatchptr=i;
			ret = &node->data;
		}
	}

	return ret;
}


void **dtrie_contains(struct dtrie_node_t *root, const char *number, const unsigned int numberlen)
{
	int nmatch;
	void **ret;
	ret = dtrie_longest_match(root, number, numberlen, &nmatch);

	if (nmatch == numberlen) return ret;
	return NULL;
}
